/* Copyright 1999 VLSI Vision Ltd. Free for use under the Gnu Public License */

/*
 	Version 0.50
	
	pport.c

	Project: Vision DUAL Camera

	interface to parallel port, takes care of reading
	writing data and the operation mode of the port
	
*/
#include <conio.h>
#include <dos.h>
#include <string.h>

#ifdef _WINDOWS
#include <windows.h>
#include <mmsystem.h>
#else
#include <time.h>
#endif


#include "vvl_defs.h"
#include "pport.h"


#define HARDWIRED		/* use hardwired port addressing on data upload */


#define WHILE_OUT(gubbins)	{ DWORD k; k=Port->Whileout;while((gubbins) && (--k > 0));\
							if(k==0) goto WhileoutError; }


/* error code messages */
char *ErrorMessagePPort[] = \
{	"OK",
	"Error in DetectPortType()",
	"Error in InitialisePPort()",
	"Error in ClosePort()",
	"Error in Attempt1284Negotiation",

	"Error in SetTransferMode",
	"Error in EndTransferMode",
	"Error in SetUploadMode",
	"Error in EndUploadMode",

	"Error in Forward2Reverse",
	"Error in Reverse2Forward",
	"Error in ECPWriteAddrByte()",
	"Error in ECPWriteBuffer()",
	"Error in ECPReadBuffer()",

	"Error in SimECPWriteAddrByte()",
	"Error in SimECPWriteBuffer()",
	"Error in SimECPReadBuffer()",

	"Error in NibbleReadBuffer()",

	"Error in ECPUpload",
	"Error in RLE_ECPUpload",

	"Error in BiDirUpload",
	"Error in RLE_BiDirUpload",

	"Error in NibbleUpload",
	"Error in RLE_NibbleUpload",
    "Opps",
    "Opps"   }; 

ERRORCHECKPPORT ErrorCheckPPort;


/*--------------------------------------------------------------------------*/
/*	DetectPortType														*/
/*--------------------------------------------------------------------------*/
int DetectPortType( BYTE PortNumber, PPORT *Port ) {
	BYTE numports;

	Port->PortType = NULL_PORT_TYPE;

	numports = ( (*((BYTE far *)MK_FP(0x40, 0x11))) &0xC0 )>>6;
/*	switch (numports)
	{
		case 0:	MessageBox((HWND)NULL,"none","Ports",MB_OK);  
				break;
		case 1:	MessageBox((HWND)NULL,"one","Ports",MB_OK);  
				break;
		case 2:	MessageBox((HWND)NULL,"two","Ports",MB_OK);  
				break;
		case 3:	MessageBox((HWND)NULL,"three","Ports",MB_OK);  
				break;
		default:MessageBox((HWND)NULL,"wild","Ports",MB_OK);  
				break;
	}	
*/
	if ( ( numports < LPT1 ) || ( numports > LPT3 ) )
	{
		ErrorCheckPPort.Code =  DETECTPORTTYPE;
		return( FALSE );
	}
	if ( ( PortNumber < LPT1 ) || ( PortNumber > numports )	)
	{
		ErrorCheckPPort.Code =  DETECTPORTTYPE;
		return( FALSE );
	}
	Port->PortNumber = PortNumber;	

	Port->BaseAddr = *((WORD far *)MK_FP(0x40, 6 + (2 * Port->PortNumber)));

	if ( Port->BaseAddr == 0) {
		ErrorCheckPPort.Code =  DETECTPORTTYPE;
		return( FALSE );
	}
	if ( (Port->BaseAddr != 0x378) && (Port->BaseAddr != 0x278) && (Port->BaseAddr != 0x3BC))
	{
		ErrorCheckPPort.Code =  DETECTPORTTYPE;
		return( FALSE );
	}
	Port->PrevData = _inp(Port->BaseAddr);
	Port->PrevControl = _inp(Port->BaseAddr+CONTROL_OFFSET);

  	Port->PortType = TestPortType( Port->BaseAddr );

	return( TRUE );
}



/*--------------------------------------------------------------------------*/
/*	InitialisePort															*/
/*--------------------------------------------------------------------------*/
int InitialisePPort( PPORT *Port ) {
	WORD Base;
	double diff;
	DWORD start, stop;

	Base = Port->BaseAddr;

	Port->DataAddr = Base;
	Port->StatusAddr = Base + STATUS_OFFSET;
	Port->ControlAddr = Base + CONTROL_OFFSET;
	Port->ECRAddr = Base + ECP_ECR;
	Port->DataFIFOAddr = Base + ECP_D_FIFO;

 	Port->PrevControl = _inp( Port->ControlAddr );
	if ( !(Port->PrevControl & DirBit) ) { /* ie port was in FORWARD */
		Port->PrevData = _inp( Port->DataAddr );
	}

	if ( Port->PortType & ECP_PORT_TYPE ) {
		Port->PrevECR = _inp( Base + ECP_ECR );
		
		SetECRMode( Port, ECP_STANDARD_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_STANDARD_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_CNFG_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_CNFG_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_CNFG_MODE );
		Port->PrevCnfgA = _inp( Base + ECP_CNFG_A );
		Port->PrevCnfgB = _inp( Base + ECP_CNFG_B );

		/* set CnfgA and CnfgB to default values here if required 	*/
		/* ... probably safer to not touch them for now !!				*/
		SetECRMode( Port, ECP_STANDARD_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_STANDARD_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_PS2_MODE );
		if ( GetECRBit( Port, ECR_mode ) !=  ECP_PS2_MODE ) goto InitialiseError; 
	}		

	Port->ControlShadow = (BYTE )(Port->PrevControl ^ CONTROL_INVERSION_MASK);

	SetCOMPATMode( Port );	

	Port->Timeout_mS = DEFAULT_TIMEOUT_MS;

	/* determine port timing for timeout value  */
	Port->Whileout = 100000;					/* calibration value */

    start = GetTime();                                        
	WHILE_OUT( (_inp( Port->StatusAddr ) | 0xFF) );		/* fastest type of WHILE_OUT useage */

WhileoutError:
    stop = GetTime();                                        

	if ( start == stop ) goto InitialiseError;	/* Err, something went wrong !!!! */

	diff = ((stop - start) * 1000)/(DWORD)CLOCKS_PER_SEC;

    Port->Whileout = ((DWORD)Port->Timeout_mS * (DWORD)100000) / (DWORD)diff;

    /* timeout value is the number of port reads done in WHILE_OUTTIME seconds */

 	return( TRUE );

InitialiseError:
	ErrorCheckPPort.Code =  INITIALISEPPORT;

	return( FALSE );
}


/*-----------------------------------------------------------------*/
/* ClosePort - Restores state of the parallel port to that before  */
/*-----------------------------------------------------------------*/
void ClosePPort( PPORT *Port ) {
	WORD Base;

	Base = Port->BaseAddr;

	if ( Port->PortType & ECP_PORT_TYPE ) {
		SetECRMode( Port, ECP_CNFG_MODE );
		_outp( Base + ECP_CNFG_A, Port->PrevCnfgA );
		_outp( Base + ECP_CNFG_B, Port->PrevCnfgB );
		_outp( Port->ECRAddr, Port->PrevECR );
	}
 
	if ( !(Port->PrevControl & DirBit) ) { /* ie port was in FORWARD */
		_outp( Port->DataAddr, Port->PrevData );
	}

	return;
}


#define VPP_WAIT_TIME	200

/*--------------------------------------------------------------------------*/
/*	SendVPP - Sends a number of data bytes to the  camera without any 		*/
/*				activity on the control lines.								*/
/*--------------------------------------------------------------------------*/
int SendVPP( PPORT *Port, BYTE Bytes, BYTE *Buffer ) {
	int i,j;

	for( i = 0; i < Bytes; i++ ) {
		for( j = 0; j < VPP_WAIT_TIME; j++ ) { /* this loop for delay purposes */
			PutData( Port, Buffer[ i ] );
		}
	}

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	GetTimeout - Allows the retrieval of the current timeout value			*/
/*--------------------------------------------------------------------------*/
int GetTimeout( PPORT *Port, WORD *Timeout_mS ) {

	*Timeout_mS = Port->Timeout_mS;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetTimeout - Allows the setting of the timeout value					*/
/*--------------------------------------------------------------------------*/
int SetTimeout(  PPORT *Port, WORD Timeout_mS ) {
	DWORD tmp;

	if ( Timeout_mS < 10 ) Timeout_mS = 10;
	if ( Timeout_mS > 1000 ) Timeout_mS = 1000;
	
	tmp = ((DWORD)Timeout_mS * Port->Whileout) / (DWORD)Port->Timeout_mS;

	if ( tmp < 100 ) tmp = 100;

	Port->Timeout_mS = Timeout_mS;
	Port->Whileout = tmp;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetTransferMode - Negotiates from compatibility mode into the requested */
/*					  1284 standard mode.									*/
/*--------------------------------------------------------------------------*/
int SetTransferMode( PPORT *Port, BYTE NewMode ) {
	BYTE request_id;

	if ( Port->TransferMode != COMPAT_TRANSFER ) goto ModeNotChanged;
	
	request_id = NewMode & REQUEST_DEVICE_ID;
	
	switch( (NewMode & ~REQUEST_DEVICE_ID) ) {

	case ECP_TRANSFER:
		if ( !Negotiate2SetupPhase( Port, (EXTEN_ECP_MODE | request_id), 0 ) ) goto ModeNotChanged;
		if ( !ECPSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	case NIBBLE_TRANSFER:
		if ( !Negotiate2SetupPhase( Port, (EXTEN_NIBBLE_MODE | request_id), 0 ) ) goto ModeNotChanged;
		if ( !NibbleSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	default:
		goto ModeNotChanged;
	}
	return( TRUE );

ModeNotChanged:
	ErrorCheckPPort.Code = SETTRANSFERMODE;
	return( FALSE );
}

/*-----------------------------------------------------------------*/
/* EndTransferMode												  */
/*-----------------------------------------------------------------*/
int EndTransferMode( PPORT *Port ) {
	if ( Port->TransferMode == COMPAT_TRANSFER ) return( TRUE );

	switch( Port->TransferMode ) {

	case ECP_TRANSFER:
		if ( Port->Direction == REVERSE_DIR ) {
			if ( !Reverse2Forward( Port ) ) goto ModeNotChanged;
		}
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	case NIBBLE_TRANSFER:
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	default:
		goto ModeNotChanged;
	}
	return( TRUE );

ModeNotChanged:
	ErrorCheckPPort.Code = ENDTRANSFERMODE;
	return( FALSE );

DoImmediateTermination:
	Immediate1284Termination( Port );
	ErrorCheckPPort.Code = ENDTRANSFERMODE;
	return( FALSE );
}



/*------------------------------------------------------------------*/
/* SetUploadMode														*/
/*------------------------------------------------------------------*/
int SetUploadMode( PPORT *Port, BYTE UploadMode, BYTE ModeFlags ) {

	if ( Port->TransferMode != COMPAT_TRANSFER ) goto ModeNotChanged;

	if ( !Negotiate2SetupPhase( Port, REQUEST_LINK, (UploadMode | ModeFlags) ) )
		goto ModeNotChanged;

	switch( UploadMode ) {
	
	case ECP_UPLOAD:  	/* since this is almost the same as ECP		*/
					  	/* it can share the ECP Setup stuff.		*/
		if ( !ECPSetupPhase( Port ) ) goto ModeNotChanged;
		if ( !Forward2Reverse( Port ) ) {
			Valid1284Termination( Port );
			goto ModeNotChanged;
		}
		break;

	case NIBBLE_UPLOAD:
		if ( !NibbleSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	default:
		goto ModeNotChanged;
	}

	Port->TransferMode = UploadMode;
	return( TRUE );

ModeNotChanged:
	ErrorCheckPPort.Code = SETUPLOADMODE;
	return( FALSE );
}



/*-----------------------------------------------------------------*/
/* EndUploadMode													  */
/*-----------------------------------------------------------------*/
int EndUploadMode( PPORT *Port ) {
	if ( Port->TransferMode == COMPAT_TRANSFER ) return( TRUE );

	switch( (Port->TransferMode & ~RLE_DATA_FLAG) ) {

	case ECP_UPLOAD:
	case BIDIR_UPLOAD:
		if ( Port->Direction == REVERSE_DIR )
			if ( !Reverse2Forward( Port ) ) goto DoImmediateTermination;
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		 	
		break;

	case NIBBLE_UPLOAD:
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	default:
		goto ModeNotChanged;
	}
	return( TRUE );

ModeNotChanged:
	ErrorCheckPPort.Code = ENDUPLOADMODE;
	return( FALSE );

DoImmediateTermination:
	Immediate1284Termination( Port );
	ErrorCheckPPort.Code = ENDUPLOADMODE;
	return( FALSE );
}

/*-----------------------------------------------------------------*/
/* Attempt1284Negotiation										  */
/*-----------------------------------------------------------------*/
int Attempt1284Negotiation( PPORT *Port ) {
	
	DWORD tmp_whileout;
	
 	PutData( Port, 0 );	/* ie request reverse nibble mode */

	SetControlBit( Port, Active_1284 );
	ClearControlBit( Port, nAutoFd );
	
	/* frig the WHILE_OUT mechanism to give a very short timeout */
	tmp_whileout = Port->Whileout;
	Port->Whileout = 50;

	WHILE_OUT( GetStatusBit( Port, nAck ) );			/* check nAck LO ...	*/
	WHILE_OUT( !((GetStatus( Port ) & (PError|nFault|XFlag))
					== (PError|nFault|XFlag))  );			/* ... and these lines go HI 	*/

	Port->Whileout = tmp_whileout;
	/* if we got here then the dongle has responded */

	ClearControlBit( Port, nStrobe );
	
	SetControlBit( Port, nInit );	/* do this to ensure that the dongle does reset	*/
	ClearControlBit( Port, Active_1284 );			
   
	/* setting Active_1284 high here will cause the dongle to abort the */
	/* negotiation while it is waiting for event (4) to occur.			*/
	/* it jumps directly to its Immmediate1284Termination routine.		*/

	/* wait for peripheral to see the termination .. */
//	WHILE_OUT( GetStatusBit( Port, XFlag ) );

	SetCOMPATMode( Port );

	/* replace correct WHILE_OUT counts */
	Port->Whileout = tmp_whileout;
 //	OutputDebugStr( "T" );
	return( TRUE );

WhileoutError:
	 SetCOMPATMode( Port );

	/* replace correct WHILE_OUT counts */
	Port->Whileout = tmp_whileout;
//	OutputDebugStr( "F" );
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	Forward2Reverse															*/
/*--------------------------------------------------------------------------*/
int Forward2Reverse( PPORT *Port ) {

	SetControlBit( Port, DirBit );		/* tri-state data bits */
	ClearControlBit( Port, HostAck );

	ClearControlBit( Port, nReverseRequest );
	WHILE_OUT( GetStatusBit( Port, nAckReverse ) );

	Port->Direction = REVERSE_DIR;
	return( TRUE );

WhileoutError:
	SetControlBit( Port, nReverseRequest );
	ClearControlBit( Port, DirBit );		/* drive data bits */
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	Reverse2Forward															*/
/*--------------------------------------------------------------------------*/
int Reverse2Forward( PPORT *Port ) {

	SetControlBit( Port, nReverseRequest );
	WHILE_OUT( !GetStatusBit( Port, nAckReverse ) );
	ClearControlBit( Port, DirBit );		/* drive data bits */

	Port->Direction = FORWARD_DIR;
	return( TRUE );

WhileoutError:
	ClearControlBit( Port, nReverseRequest );
	SetControlBit( Port, DirBit );		/* tri-state data bits */
	return( FALSE );
}




/*--------------------------------------------------------------------------*/
/*	ECPWriteBuffer															*/
/*--------------------------------------------------------------------------*/
int ECPWriteBuffer( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	WORD j;
	BYTE __huge *ptr;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 
		   || (Port->TransferMode != ECP_TRANSFER) 
		   || (Port->Direction != FORWARD_DIR) ) {
		
		ErrorCheckPPort.Code =  ECPWRITEBUFFER;
		return( FALSE );	
	}

	EmptyFIFO( Port );

	SetControlBit( Port, HostAck );	  				/* NOT what I'd expect to do !!! */
	SetControlBit( Port, HostClk ); 	 		   
	SetECRMode( Port, ECP_ECP_MODE );				

	ptr = Data;

	for( i = 0; i < (Bytes/ECP_FIFO_SIZE); i++ ) {	

		WHILE_OUT( !GetECRBit( Port, ECR_empty ) );  	/* wait for FIFO to empty */
		
		/* FIFO is now empty so safe to write at least ECP_FIFO_SIZE BYTE s */
		for( j = 0; j < ECP_FIFO_SIZE; j++ ) {
			_outp( Port->DataFIFOAddr, *ptr++);
		}
	}

	WHILE_OUT( !(GetECRBit( Port, ECR_empty )) ); 	
	for( j = 0; j < (Bytes%ECP_FIFO_SIZE); j++ ) {
		_outp( Port->DataFIFOAddr, *ptr++);
	}

	/* wait for last BYTE to be sent */
	WHILE_OUT( !(GetECRBit( Port, ECR_empty )) ); 	
	SetECRMode( Port, ECP_PS2_MODE );				

	return( TRUE );

WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	EmptyFIFO( Port );
	ErrorCheckPPort.Code =  ECPWRITEBUFFER;
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	ECPReadBuffer														*/
/*--------------------------------------------------------------------------*/
int ECPReadBuffer( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	WORD j;
	BYTE __huge *ptr;
	BYTE __huge *end_ptr;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 	 
		   || ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD)) 
		   || (Port->Direction != REVERSE_DIR) ) {
		
		ErrorCheckPPort.Code =  ECPWRITEBUFFER;
		return( FALSE );	
	}
	ptr = Data;
	end_ptr = ptr + Bytes;

	if ( Bytes < ECP_FIFO_SIZE*4 ) goto FinishManual;

	ClearControlBit( Port, HostAck );	  				

	/* ECP hardware only detects the HI->LO transition on PeriphClk, therefore
	   it will have missed this transition for the first BYTE ( which the dongle
	   sends straight after doing Forward2Reverse() )
	   So here we manually read the first BYTE of data before enabling ECP hardware */ 
	WHILE_OUT( GetStatusBit( Port, PeriphClk ) );  	
	*ptr++ = GetData( Port );
	Bytes--;

	/* acknowledge for this BYTE */
	SetControlBit( Port, HostAck );	  				
	WHILE_OUT( !GetStatusBit( Port, PeriphClk ) ); 
	/* note - switching on ECP hardware will complete handshake for this BYTE by 	*/
	/* taking HostAck LO again.														*/

	/* OK we've got the first BYTE ... now let the ECP hardware do its thing */
	SetControlBit( Port, HostClk ); 	 			
	SetECRMode( Port, ECP_ECP_MODE );				

	if ( Port->BaseAddr == LPT1_BASE_ADDR )	{
	 	for( i = 0; i < (Bytes/ECP_FIFO_SIZE)-1; i++ ) {	
			WHILE_OUT( !(_inp(ECR_LPT1) & ECR_full) );   /* wait for FIFO to fill */
			for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = _inp( DATA_FIFO_LPT1 );
		}
		/* switch off automatic filling of the FIFO		 	*/
		ClearControlBit( Port, HostAck );	  				
	 	/* read any remaining bytes out the FIFO 			*/
		while( (!(_inp(ECR_LPT1) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = _inp( DATA_FIFO_LPT1 );
	}
	else {
		if ( Port->BaseAddr == LPT2_BASE_ADDR ) {
		 	for( i = 0; i < (Bytes/ECP_FIFO_SIZE)-1; i++ ) {	
				WHILE_OUT( !(_inp(ECR_LPT2) & ECR_full) );   /* wait for FIFO to fill */
				for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = _inp( DATA_FIFO_LPT2 );
			}
			ClearControlBit( Port, HostAck );	  				
			while( (!(_inp(ECR_LPT2) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = _inp( DATA_FIFO_LPT2 );
		}
		else {
			if ( Port->BaseAddr == LPT3_BASE_ADDR )	 {
			 	for( i = 0; i < (Bytes/ECP_FIFO_SIZE)-1; i++ ) {	
					WHILE_OUT( !(_inp( ECR_LPT3 ) & ECR_full) );   /* wait for FIFO to fill */
					for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = _inp( DATA_FIFO_LPT3 );
				}
				ClearControlBit( Port, HostAck );	  				
				while( (!(_inp(ECR_LPT2) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = _inp( DATA_FIFO_LPT2 );
			}
		}
	}

	while( !GetECRBit( Port, ECR_empty ) && ( ptr != end_ptr ) ) {
  		*ptr++ = _inp( Port->DataFIFOAddr );
	}

 	SetECRMode( Port, ECP_PS2_MODE );				
//	SetControlBit( Port, HostAck );

	/* now manually pull through any remaining bytes 	*/

FinishManual:
	SimECPReadBuffer( Port, ptr, (end_ptr - ptr) );
	 
	return( TRUE );

WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	EmptyFIFO( Port );
	ErrorCheckPPort.Code =  ECPREADBUFFER;
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	SimECPWriteBuffer														*/
/*--------------------------------------------------------------------------*/
int SimECPWriteBuffer( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	BYTE __huge *ptr;

	if ( (Port->TransferMode != ECP_TRANSFER)
			|| (Port->Direction != FORWARD_DIR) ) {
	   ErrorCheckPPort.Code =  SIMECPWRITEBUFFER;
	   return( FALSE );	
	}

	SetControlBit( Port, HostAck );	  	/* this is a data transfer	*/

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {	
		PutData( Port, *ptr++ );

		ClearControlBit( Port, HostClk );
		WHILE_OUT( !GetStatusBit( Port, PeriphAck ) );  	

		SetControlBit( Port, HostClk );
		WHILE_OUT( GetStatusBit( Port, PeriphAck ) );  	
	}

	return( TRUE );

WhileoutError:
	SetControlBit( Port, HostClk );
	ErrorCheckPPort.Code =  SIMECPWRITEBUFFER;
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	SimECPReadBuffer														*/
/*--------------------------------------------------------------------------*/
int SimECPReadBuffer( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	BYTE __huge *ptr;

	if ( ( !(Port->PortType & BIDIR_PORT_TYPE))
			|| ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD))
			|| (Port->Direction != REVERSE_DIR) ) {
	   ErrorCheckPPort.Code =  SIMECPREADBUFFER;
	   return( FALSE );	
	}

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {	
		ClearControlBit( Port, HostAck );
 		WHILE_OUT( GetStatusBit( Port, PeriphClk ) );

		*ptr++ = GetData( Port );		/* read data on rising edge of PeriphClk */

		SetControlBit( Port, HostAck );
		WHILE_OUT( !GetStatusBit( Port, PeriphClk ) );
  	}
		 
	return( TRUE );

WhileoutError:
	ClearControlBit( Port, HostAck );
	ErrorCheckPPort.Code =  SIMECPREADBUFFER;
	return( FALSE );
}

#define UPPER_NIBBLE	0xF0
#define LOWER_NIBBLE	0x0F


BYTE UnscrambleNibble( BYTE Data ) {

	Data = Data & NIBBLE_DATA_BITS;

	if ( Data & Busy ) {
		return( ((Busy>>1) | (Data & 0x78))>>3 );
	}
	else {
		return( Data>>3);
	   	/* busy (bit 7 ) was not set, so no action is required ( bit 6 is already zero ) */
	}
}



/*--------------------------------------------------------------------------*/
/*	NibbleReadBuffer														*/
/*--------------------------------------------------------------------------*/
int NibbleReadBuffer( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {

	DWORD i;
	BYTE __huge *ptr;
	WORD tmp;

	if ( Port->TransferMode != NIBBLE_TRANSFER ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
 //		WHILE_OUT( GetStatusBit( Port, nDataAvail ) );
		
		ClearControlBit( Port, HostBusy );
		WHILE_OUT( GetStatusBit( Port, PtrClk ) );
		tmp = UnscrambleNibble( GetStatus( Port ) );
		
		/* now invert and shift BUSY bit */

		SetControlBit( Port, HostBusy );
		WHILE_OUT( !GetStatusBit( Port, PtrClk ) );

		ClearControlBit( Port, HostBusy );
		WHILE_OUT( GetStatusBit( Port, PtrClk ) );
		*ptr++ = (BYTE )(tmp | (UnscrambleNibble(GetStatus( Port ))<<4) );

		SetControlBit( Port, HostBusy );
		WHILE_OUT( !GetStatusBit( Port, PtrClk ) );
	}
	return( TRUE );

WhileoutError:

	SetControlBit( Port, HostBusy );
	ErrorCheckPPort.Code =  NIBBLEREADBUFFER;
	return( FALSE );
}





/****************************************************************************/
/*																			*/
/*	Special PPC Upload Functions	 										*/
/*																			*/
/*																			*/
/****************************************************************************/


/*--------------------------------------------------------------------------*/
/*	ECPUpload																*/
/*--------------------------------------------------------------------------*/
int ECPUpload( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {

	if ( (!(Port->PortType & ECP_PORT_TYPE) ) 
			|| (Port->TransferMode != ECP_UPLOAD) ) {
		ErrorCheckPPort.Code =  ECPUPLOAD;
		return( FALSE );
	}

	return( ECPReadBuffer( Port, Data, Bytes ) );
}


/*--------------------------------------------------------------------------*/
/*	NibbleUpload															*/
/*--------------------------------------------------------------------------*/
int NibbleUpload( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
    int rtn;
    
	rtn = FALSE;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	if ( Port->BaseAddr == LPT1_BASE_ADDR )
		rtn = LPT1_NibbleUpload( Port, Data, Bytes );
	else
		if ( Port->BaseAddr == LPT2_BASE_ADDR )
			rtn = LPT2_NibbleUpload( Port, Data, Bytes );
		else
			if ( Port->BaseAddr == LPT3_BASE_ADDR )
				rtn = LPT3_NibbleUpload( Port, Data, Bytes );

	if ( rtn == FALSE )	{
		ErrorCheckPPort.Code =  NIBBLEUPLOAD;
	}

	return( rtn );
}


/****************************************************************************/
/*																			*/
/*	Private functions of PPORT Module below here 							*/
/*																			*/
/****************************************************************************/


void SetCOMPATMode( PPORT *Port ) {

	PutControl( Port, (nStrobe|nAutoFd|nInit) );

	PutData( Port, 0xFF ); 	/* just in case peripheral is still driving lines */

	SetDataDirection( Port, FORWARD_DIR );

	Port->TransferMode = COMPAT_TRANSFER;
	Port->Direction = FORWARD_DIR;

	return;
}




/*--------------------------------------------------------------------------*/
/*	Negotiate2SetupPhase													*/
/*--------------------------------------------------------------------------*/
int Negotiate2SetupPhase( PPORT *Port, BYTE ExtensibilityByte, BYTE LinkMode ) {

	PutData( Port, ExtensibilityByte );

	SetControlBit( Port, Active_1284 );
	ClearControlBit( Port, nAutoFd );

 	WHILE_OUT( GetStatusBit( Port, nAck ) );			/* check nAck LO ...	*/
	WHILE_OUT( !((GetStatus( Port ) & (PError|nFault|XFlag))
				== (PError|nFault|XFlag))  );			/* ... and these lines go HI 	*/

	/* ok we have got a 1284 peripheral attached */

	/* the following LO going pulse on nStrobe is now ignored by the 		*/
	/* periperal which will read the ExtensiblityByte as soon as it sees	*/
	/* Active_1284 going HI.												*/
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );

 	SetControlBit( Port, (nAutoFd | nStrobe) );		   /* min 0.5uS LO pulse required 	*/
  
  	WHILE_OUT( !GetStatusBit( Port, nAck ) ); 					//  -- (event 6)		
	
//	ExtensibilityByte &= ~REQUEST_DEVICE_ID;		// clear Device ID Request bit

	if ( ExtensibilityByte == EXTEN_NIBBLE_MODE ) {
		if ( GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}
	else {
		if ( !GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}

	if ( ExtensibilityByte & REQUEST_LINK ) {
		/* an Extensiblity Link was requested */

		PutData( Port, LinkMode );								
					
		ClearControlBit( Port, nStrobe);							//  -- (event 51)
	  	WHILE_OUT( GetStatusBit( Port, nAck ) ); 							

		SetControlBit( Port, (nAutoFd | nStrobe) );		   			//  -- (event 53)
	  	WHILE_OUT( !GetStatusBit( Port, nAck ) ); 				   		

		if ( !GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}
  	/* negotiation phase complete */

	return( TRUE );	

WhileoutError:
	SetCOMPATMode( Port );
 	return( FALSE );

ModeNotSupported:
	SetCOMPATMode( Port );
 	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	Valid1284Termination													*/
/*--------------------------------------------------------------------------*/
int Valid1284Termination( PPORT *Port ) {

	SetControlBit( Port, nInit );	/* do this to ensure that the dongle 	*/
									/* doesn't get a reset.					*/
									/* Reset is  nInit = Active_1284 = LO	*/

	SetControlBit( Port, nAutoFd );

	ClearControlBit( Port, Active_1284 );			

//
//	WHILE_OUT( GetStatusBit( Port, (Busy|nFault) ) 	
//						!= (Busy|nFault) );			/* 		-- (event 23)	*/

                                                
	WHILE_OUT( GetStatusBit( Port, nAck ) );		/*		-- (event 24)	*/ 
	ClearControlBit( Port, nAutoFd );				/*		-- (event 25)	*/
                                                
                                                
	WHILE_OUT( !GetStatusBit( Port, nAck ) );		/* 		-- (event 27)	*/

	SetControlBit( Port, nAutoFd );

	SetCOMPATMode( Port );

	return( TRUE );

WhileoutError:
	Immediate1284Termination( Port );
	return( FALSE );
}




/*--------------------------------------------------------------------------*/
/*	Immmediate1284Termination												*/
/*--------------------------------------------------------------------------*/
void Immediate1284Termination( PPORT *Port ) {

	SetControlBit( Port, nInit );	/* make sure dongle will not be reset	*/
	ClearControlBit( Port, Active_1284 );

	/* wait for peripheral to see the termination .. */
	WHILE_OUT( GetStatusBit( Port, XFlag ) );

	SetCOMPATMode( Port );
	return;

WhileoutError:
   	SetCOMPATMode( Port );
	return;
}




											
/*--------------------------------------------------------------------------*/
/*	ECPSetupPhase															*/
/*--------------------------------------------------------------------------*/
int  ECPSetupPhase( PPORT *Port ) {
 
 	/* Assertion - negotiation phase complete before this functions is 		*/
 	/* called ... now do ECP setup phase.									*/

	ClearControlBit( Port, HostAck );

	WHILE_OUT( !GetStatusBit( Port, nAckReverse ) );

	EmptyFIFO( Port );

	Port->TransferMode = ECP_TRANSFER;
	Port->Direction = FORWARD_DIR;
 	return( TRUE );

WhileoutError:
	SetCOMPATMode( Port );
 	return( FALSE );
}

 

/*--------------------------------------------------------------------------*/
/*	AttemptForward2Reverse	  												*/
/*--------------------------------------------------------------------------*/
int AttemptForward2Reverse( PPORT *Port ) {
	static int WaitingForReverseAck = FALSE;

	if ( Port->Direction != FORWARD_DIR ) {
		WaitingForReverseAck = FALSE;
		return( FALSE );
	}
	
	if ( !WaitingForReverseAck ) {
		SetControlBit( Port, DirBit );		/* tri-state data bits */
		ClearControlBit( Port, HostAck );
		ClearControlBit( Port, nReverseRequest );

		/* wait a bit here */
		ClearControlBit( Port, nReverseRequest );
		ClearControlBit( Port, nReverseRequest );
		ClearControlBit( Port, nReverseRequest );

		WaitingForReverseAck = TRUE;	
	}

	if ( !GetStatusBit( Port, nAckReverse ) ) { 	/* peripheral has acknowledged */
		Port->Direction = REVERSE_DIR;
		WaitingForReverseAck = FALSE;
		return( TRUE );
	}	
	else {
		return( FALSE );							/* peripheral has not acknowledged */
	}
}





/*--------------------------------------------------------------------------*/
/*	NibbleSetupPhase						 								*/
/*--------------------------------------------------------------------------*/
int  NibbleSetupPhase( PPORT *Port ) {
 
 	/* Assertion - negotiation phase complete before this functions is 		*/
 	/* called ... now do Nibble setup phase. 								*/

	SetDataDirection( Port, REVERSE_DIR );

//	ClearControlBit( Port, HostBusy );

	Port->TransferMode = NIBBLE_TRANSFER;
	Port->Direction = REVERSE_DIR;

 	return( TRUE );
}



/* ====================================================================	*/
/*																		*/
/* General routines										*/
/*																		*/
/* ====================================================================	*/

WORD TestPortType( WORD BaseAddr ) {
	WORD port_type;

	port_type = COMPAT_PORT_TYPE;
	
	if ( TestGenericECP( BaseAddr ) ) 
		port_type |= (ECP_PORT_TYPE | BIDIR_PORT_TYPE);

	return( port_type );
}


int TestGenericECP( WORD BaseAddr ) {
	BYTE old_ecr, ecr;

	old_ecr = _inp( BaseAddr + ECP_ECR );
	ecr = (old_ecr & 0x1F) | ECP_TEST_MODE;	
	_outp( BaseAddr + ECP_ECR, ecr );
	
	/* check that the mode switch did take */
	if ( _inp( BaseAddr + ECP_ECR ) != ecr ) goto Fail_ECP_Test;
	   
	_outp( BaseAddr + ECP_T_FIFO, 'H' );
	_outp( BaseAddr + ECP_T_FIFO, 'i' );
	if ( _inp( BaseAddr + ECP_T_FIFO ) != 'H' ) goto Fail_ECP_Test;
	if ( _inp( BaseAddr + ECP_T_FIFO ) != 'i' ) goto Fail_ECP_Test;

	/* OK it seems to be an ECP port */
	_outp( BaseAddr + ECP_ECR, old_ecr ); /* restore old ECR */

	return( TRUE );

	
Fail_ECP_Test:
	_outp( BaseAddr + ECP_ECR, old_ecr ); /* restore old ECR */
	return( FALSE );		 
}

 
void SetECRMode( PPORT *Port, BYTE Mode ) {
	_outp( Port->ECRAddr, ( Mode | ECR_nErrIntrEn | ECR_serviceIntr ) );
}

BYTE GetECRBit( PPORT *Port, BYTE BitMask ) {
	BYTE tmp;
	tmp = _inp( Port->BaseAddr + ECP_ECR );
	return( tmp & BitMask );
}

void EmptyFIFO( PPORT *Port ) {
	if ( Port->PortType & ECP_PORT_TYPE ) {
		while( !GetECRBit( Port, ECR_empty ) ) _inp( Port->DataFIFOAddr );
	}
	return;
}



/* ====================================================================	*/
/*																		*/
/* Low level port access routines										*/
/*																		*/
/* ====================================================================	*/
void SetDataDirection( PPORT *Port, BYTE NewDirection ) {

	if ( NewDirection == Port->Direction ) return;	/* requested == current */

	if ( NewDirection == FORWARD_DIR ) {
		ClearControlBit( Port, DirBit );		
		Port->Direction = FORWARD_DIR;
	}
	else {
		SetControlBit( Port, DirBit );
		PutData( Port, 0xFF ); 			/* for pseudo bi-dir ports */
		Port->Direction = REVERSE_DIR;
	}
	return;
}

void PutData( PPORT *Port, BYTE Data ) {
	_outp( Port->DataAddr, Data );
}

BYTE GetData( PPORT *Port ) {
	BYTE tmp;
	tmp = (BYTE )_inp( Port->DataAddr );
	return( tmp  );
}

BYTE GetStatus( PPORT *Port ) {
	return( (BYTE )(_inp( Port->StatusAddr )^ STATUS_INVERSION_MASK) );
}

BYTE GetControl( PPORT *Port ) {
	return( Port->ControlShadow );
}

void PutControl( PPORT *Port, BYTE Data ) {
	BYTE Data2;
	Data2 = (Data ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data2 );
    Port->ControlShadow = Data; 
}

BYTE GetStatusBit( PPORT *Port, BYTE BitMask ) {
	return( (BYTE )((_inp( Port->StatusAddr )^ STATUS_INVERSION_MASK) & BitMask) );
}

void SetControlBit( PPORT *Port, BYTE BitMask ) {
	BYTE Data;
	Data = ((Port->ControlShadow |= BitMask) ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data );
}	

void ClearControlBit( PPORT *Port, BYTE BitMask ) {
	BYTE Data;
	Data = ((Port->ControlShadow &= ~BitMask) ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data );
}






__inline BYTE UnscrambleUpload( BYTE Data ) {

	if ( Data & Busy ) {
		return( (Data&0x38)>>3 );
	}
	else {
		return( ((0x40)|(Data&0x38))>>3 );
	   	/* busy (bit 7 ) was not set, so inverted it and return */
	}
}




int LPT1_NibbleUpload( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	BYTE __huge *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT1, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT1 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );
 
 		WRITECTRL( CONTROL_LPT1, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT1 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}
	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
	return( FALSE );              
}


int LPT2_NibbleUpload( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	BYTE __huge *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT2, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT2 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );

		WRITECTRL( CONTROL_LPT2, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT2 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}
	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
	return( FALSE );              
}

int LPT3_NibbleUpload( PPORT *Port, BYTE __huge *Data, DWORD Bytes ) {
	DWORD i;
	BYTE __huge *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT3, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT3 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );

		WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT3 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}

	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
	return( FALSE );              
}


