/* Copyright 1999 VLSI Vision Ltd. Free for use under the Gnu Public License */

#include <conio.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <time.h>

#include "vvl_defs.h"
#include "pport.h"

#include "cmd.h"
#include "cpialink.h"
#include "cpia_cmd.h"


BOOL CheckForCPIA( BYTE PortNumber );
BOOL DoSystemTest( void );
BOOL UploadTestData( WORD DummyBlocks );
BOOL TestEEPROM( void );
void Pause( WORD mSecs );


void UploadAndSaveDRAM( void );
void UploadBlock( void );
void DoUploadImage( void );


#define STREAM_BUFFER_LEN	((DWORD)256*1024)


BYTE gPortNumber;
BYTE  __huge *gStreamBuffer;

BOOL gAEC = TRUE;

BYTE gFW_Ver, gFW_Rev, gVC_Ver, gVC_Rev;

void main( int argc, char *argv[] ) {
	int tmp, ch;
	DWORD dw_tmp;
	DWORD actual_bytes;
	BOOL rtn;

	BYTE sys, vp, stream, fatal, cmd, vp_stat;
	BYTE cam_setup0, cam_status;

	gPortNumber = 1;

	gStreamBuffer = _halloc( STREAM_BUFFER_LEN, 1 );	
	if ( gStreamBuffer == NULL ) {
		printf( "pp_test: unable to halloc() %u k bytes\n", STREAM_BUFFER_LEN/1024 );
		printf( "pp_test: Exiting ... \n" );
		exit( 1 );
	}

	if ( argv[1][0] == '-' ) {
  		if (argv[1][1] == 'p') {
         	tmp = sscanf( &argv[1][2], "%d", &gPortNumber );
    	    if ( tmp != 1 ) {
		    	printf( "pp_test: pp_test [-p<PortNumber>]\n" );
    			exit( 1 );
         	}
 		}
	}

	printf("\n\n   ======   PP_Test Ver 1-06  =====\n\n", gPortNumber );					

	if ( !CLINK_Init() ) {
		printf( "PP_TEST: CLINK_Init() failed - exiting\n" );
		exit( 1 );
	}

	if ( !CheckForCPIA( gPortNumber ) ) {
    	printf( "pp_test: unable to find DUAL on LPT%d\n", gPortNumber );
		if ( gPortNumber == 1 ) gPortNumber = 2;
		else gPortNumber = 1;
 		if ( !CheckForCPIA( gPortNumber ) )
	    	printf( "pp_test: unable to find DUAL on LPT%d\n", gPortNumber );
	}

 	do {
MainLoop:
		printf("\n   ==========   PP_Test Ver 1-06  =========\n\n", gPortNumber );					
		
		rtn = CPIA_GetCameraStatus( &sys, &vp, &stream, &fatal, &cmd, NULL, &vp_stat, NULL ); 
		if ( !rtn ) {
			printf( "Unable to GetCameraStatus\n" );
		}
		else {
			printf( "    System=0x%02x         Cmd=0x%02x       Fatal=0x%02x\n", sys, cmd, fatal );
 			if ( sys == HI_POWER_STATE ) {
				printf( "    VPCtrl=0x%02x      VPStat=0x%02x      Stream=0x%02x\n", vp, vp_stat, stream );
				rtn = CPIA_ReadVPRegs( 56, 57, 0, 0, &cam_status, &cam_setup0, NULL, NULL );
				if ( rtn ) {
					printf( "\n   CamStat=0x%02x   CamSetup0=0x%02x\n", cam_status, cam_setup0 );
				}
			}
		}


		printf( "\n   1. Get Version\n" );
 		printf( "   2. System test\n" );
		printf( "   3. Goto Suspend\n" );
		printf( "   4. Goto LoPower\n" );
 		printf( "   5. Goto HiPower\n" );
		printf( "   6. Goto Pass-Thru\n" );
		printf( "   7. Wake from Pass-Thru\n" );
		printf( "   8. Force Reset\n" );
		printf( "   9. Continuous Grab/Upload\n" );
		printf( "   0. Quit\n\n" );
		do {
			ch = getchar();
			if ( ch == '\n' ) goto MainLoop;
		} while( !isalnum( ch ) );
		ch = toupper( ch );

		switch( ch ) {
		
		case '1':
			rtn = CPIA_GetVersion( &gFW_Ver, &gFW_Rev, &gVC_Ver, &gVC_Rev );
		 	if ( rtn ) {
				printf( "\nFirmware = v%d-%02d        ", gFW_Ver, gFW_Rev );	
				printf( "Video Compressor = v%d-%02d\n\n", gVC_Ver, gVC_Rev );	
			}
			while( !_kbhit() );
			break;
		case '2':
			rtn = DoSystemTest();
			if ( !rtn ) printf( "\nSystem test = FAILED\n\n" );
			else printf( "\nSystem test = PASSED\n\n" ); 
			while( !_kbhit() );
			break;
		case '3':
			rtn = CPIA_GotoSuspend();
			if ( !rtn ) printf( "Failed to send GotoSuspend command ...\n\n" );
			else printf( "\nGotoSuspend command sent OK ...\n\n" ); 
			while( !_kbhit() );
			break;
		case '4':
			CPIA_GotoLoPower();
			break;
		case '5':
			CPIA_GotoHiPower();
			break;
		case '6':
			CPIA_WriteVCReg( 0, 0x7F, 0x80 );
			CPIA_GotoPassThrough();
			break;
		case '7':
			CLINK_SendWakeup();
			break;
		case '8':
			CLINK_ForceWatchdogReset();
			break;
		case '9':
			CPIA_GotoHiPower();
			do {
				printf( "." );
 				rtn = CPIA_GrabFrame( 0 );
				if ( !rtn ) {
					printf( "CPIA_GrabFrame failed\n" );
					break;
				}

 				do {
					rtn = CPIA_GetCameraStatus( NULL, NULL, &stream, NULL, NULL, NULL, NULL, NULL );
 					if ( !rtn ) {
						printf( "CPIA_GetCameraStatus failed\n" );
						break;
					}
 				} while( (stream != STREAM_READY) && !_kbhit() ); 

				rtn = CPIA_UploadStreamData( STREAM_BUFFER_LEN, gStreamBuffer, &actual_bytes, 1024 );
				if ( !rtn ) {
					printf( "CPIA_UploadStreamData failed\n" );
					break;
				}
				else printf( "|" );
			} while( !_kbhit() );
			CPIA_GotoLoPower();
			break;
		}

   	} while( ch != '0' );
 
	CPIA_CmdFini();

	CLINK_Fini();

 	exit( 0 );
}

BOOL CheckForCPIA( BYTE PortNumber ) {
	BOOL rtn;
	WORD base;
	BYTE type;
	char deviceID[ 256 ];

	if ( !CLINK_DetectPort( PortNumber, &base, &type ) ) { 
		printf( "\npp_test: LPT%d does not exist\n", PortNumber );
		return( FALSE );
	}
	else {
		printf( "\nLPT%d at 0x%x -- ", PortNumber, base );
		if ( type & ECP_PORT_TYPE ) printf( "ECP / " );
		if ( type & COMPAT_PORT_TYPE ) printf( "4 Bit\n" );
		printf( "\n" );
	 }

	if ( !CLINK_SelectPort( PortNumber ) ) {
		printf( "pp_test: SelectPort() failed\n" );
		return( FALSE );
	}

	if ( !CLINK_SetPortInfo( type, 800 ) ) {
		printf( "pp_test: GetPortInfo() failed\n" );
		return( FALSE );
	}

	if ( !CLINK_SendWakeup() ) {
		printf( "pp_test: SendWakeup() failed\n" );
		return( FALSE );
	}

	if ( !CLINK_Get1284DeviceID( deviceID, 256 ) ) {
		printf( "pp_test: Get1284DeviceID() failed\n" );
		return( FALSE );
	}
	printf( "DeviceID = %s\n", deviceID );	

	if ( strstr( deviceID, "VLSI Vision Ltd" ) == NULL ) {
		printf( "pp_test: DeviceID did not contain 'VLSI Vision Ltd'\n" );
		return( FALSE );
	}

	rtn = CPIA_GetCameraStatus( NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );

	return( rtn );
}


BOOL DoSystemTest( void ) {
	BOOL rtn, failed;
	BYTE sys, vp, stream, fatal, cmd, vp_stat;
	BYTE cam_setup0, cam_status;
	WORD base, timeout;
	BYTE port_type;
	BYTE devicehi, devicelo;
	BOOL I2CReadCmdAvailable;
		
	failed = FALSE;

	printf( "Getting Firmware / VC version info   --- " );
	rtn = CPIA_GetVersion( &gFW_Ver, &gFW_Rev, &gVC_Ver, &gVC_Rev );
 	if ( rtn ) {
		printf( "OK\n" );
		printf( "  - Firmware = v%d-%02d    ", gFW_Ver, gFW_Rev );	
		printf( "  VC = v%d-%02d\n", gVC_Ver, gVC_Rev );	
		if ( ((gFW_Ver==1)&&(gFW_Rev > 8)) || (gFW_Ver > 1) ) 
			I2CReadCmdAvailable = TRUE;
		else
			I2CReadCmdAvailable = FALSE;
	}
	else {
		printf( "FAILED\n" );
		printf( "    Aborting System Test\n" );
		goto SystemTestFailed;
	}

	printf( "Clearing Fatal and Cmd status        --- " );
	rtn = CPIA_ModifyCameraStatus( 3, 0, 0 ); 
	rtn &= CPIA_ModifyCameraStatus( 4, 0, 0 ); 
	if ( !rtn ) goto SendFailed;
	else printf( "OK\n" );

	printf( "Switching to Hi-Power mode           --- " );

	rtn = CPIA_GotoHiPower();
	rtn &= CPIA_GetCameraStatus( &sys, &vp, &stream, &fatal, &cmd, NULL, &vp_stat, NULL ); 
	if ( rtn && (!cmd) && (!fatal) ) printf( "OK\n" );
	else {
		printf( "FAILED\n" );
		if ( fatal & VP_CTRL_FLAG ) printf( "    VP_CTRL FatalError\n" );
		if ( fatal & CAPTURE_FLAG ) printf( "    CAPTURE FatalError\n" );
		if ( cmd & VP_CTRL_FLAG ) printf( "    VP_CTRL CmdError\n" );
		if ( cmd & CAPTURE_FLAG ) printf( "    CAPTURE CmdError\n" );
		printf( "    Aborting System Test\n" );
		goto SystemTestFailed;					
	}

	CPIA_WriteVCReg( 0, 0x7F, 0x80 );

  	/* test dummy data transfer in 4 bit mode */
	rtn = CLINK_GetPortInfo( &base, &port_type, &timeout );
	rtn = CLINK_SetPortInfo( (port_type&COMPAT_PORT_TYPE), timeout );
	printf( "Uploading test data in Nibble mode   --- " );
	rtn = UploadTestData( 256 );
	if ( rtn ) printf( "OK\n" );
	else printf( "FAILED\n" );

  	/* test dummy data transfer in ECP mode if possible*/
	if ( port_type & ECP_PORT_TYPE ) {
		rtn = CLINK_SetPortInfo( port_type, timeout );
		printf( "Uploading test data in ECP mode      --- " );
		rtn = UploadTestData( 256 );
		if ( rtn ) printf( "OK\n" );
		else printf( "FAILED\n" );
	}
	else {
		printf( "Unable to upload test data in ECP mode\n" );
	}

	/* disable AEC, ACB and wait */
	printf( "Checking for camera ...\n" );
	rtn = CPIA_SetColourBalance( 3, 0, 0, 0 );
	rtn = CPIA_SetExposure( 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	Pause( 400 );
	CPIA_ReadVPRegs( 56, 57, 0, 0, &cam_status, &cam_setup0, NULL, NULL );
	if ( !(cam_status & 0x20) ) { 
		printf( "  - Camera not detected by VP        --- FAILED\n" );
		failed = TRUE;
	}

	/* check VP_Busy --> no camera OR unprogrammed EEPROM*/
	if ( cam_status & 0x40 ) {
		printf( "  - VPBusy=HI, no camera/bad eeprom  --- FAILED\n" );
		failed = TRUE;
	}
  	
	/* write camera register */	
	printf( "  - I2C write to sensor              --- " );
	rtn = CPIA_ModifyCameraStatus( 4, 0, 0 );	/* clear the CmdError field */
	rtn &= CPIA_SerialStart();
	rtn &= CPIA_SerialWrite( 0x20 );
	rtn &= CPIA_SerialWrite( 0x80 );	/* MSB set for auto-inc */
	rtn &= CPIA_SerialStop();
	rtn &= CPIA_GetCameraStatus( NULL, NULL, NULL, NULL, &cmd, NULL, NULL, NULL );
	if ( !rtn || (cmd&SYSTEM_FLAG) ) {
		printf( "FAILED\n" );
		failed = TRUE;
	}
	else {
		printf( "OK\n" );
	}

	if ( I2CReadCmdAvailable ) {
		printf( "  - I2C read from sensor             --- " );
		rtn = CPIA_SerialStart();
		rtn &= CPIA_SerialWrite( 0x21 );
		rtn &= CPIA_SerialRead( TRUE, &devicehi );	/* dummy read to get rid of index */
		rtn &= CPIA_SerialRead( TRUE, &devicehi );
		rtn &= CPIA_SerialRead( FALSE, &devicelo );
		rtn &= CPIA_SerialStop();
		if ( !rtn || (cmd&SYSTEM_FLAG) ) {
			printf( "FAILED\n" );
			failed = TRUE;
		}
		else {
			WORD id;
			printf( "OK\n" );
			id = (devicehi<<4) + (devicelo>>4);
			printf( "  - Sensor ID = %d  Mask Rev = %d\n", id, (devicelo&0x0F) );
		}
	}
	else {
		printf( "  - Firmware does not support I2C read\n" );
	} 

	/* read EEPROM and verify checksum */
	printf( "Testing I2C to EEPROM ...\n" );
	printf( "  - I2C write to EEPROM              --- " );
	rtn = CPIA_ModifyCameraStatus( 4, 0, 0 );	/* clear the CmdError field */
	rtn &= CPIA_SerialStart();
	rtn &= CPIA_SerialWrite( 0xA0 );
	rtn &= CPIA_SerialWrite( 0x00 );
	rtn &= CPIA_SerialStop();
	rtn &= CPIA_GetCameraStatus( NULL, NULL, NULL, NULL, &cmd, NULL, NULL, NULL );
	if ( !rtn || (cmd&SYSTEM_FLAG) ) {
		printf( "FAILED\n" );
		failed = TRUE;
	}
	else {
	 	printf( "OK\n" );
	}		

	if ( I2CReadCmdAvailable ) {
		printf( "  - Verifying EEPROM checksum        --- " );
		rtn = TestEEPROM();
		if ( rtn ) printf( "OK\n" );
		else {
			printf( "FAILED\n" );
			failed = TRUE;
		}
	}
	else {
		printf( "  - Firmware does not support I2C read\n" );
	} 

	CPIA_GotoLoPower();

	if ( !failed ) return( TRUE );
	else return( FALSE );

SystemTestFailed:
	CPIA_GotoLoPower();
	return( FALSE );

SendFailed:
	printf( "\nUnable to send command --- aborting test\n\n" );
	return( FALSE );
}


void UploadAndSaveDRAM( void ) {

	FILE *file;
	int tmp;
	char fs_filename[] = "fs.txt";
	char rlb_filename[] = "rlb.txt";
	DWORD pos;

	file = fopen( fs_filename, "wb" );
	if ( file == NULL ) {
		printf( "UploadAndSaveDRAM: Unable to open %s file\n", fs_filename );
		return;
	}
	printf( "UploadAndSaveDRAM: FS save file %s\n", fs_filename );

	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) gStreamBuffer[ pos ] = 0xFE;

	if ( !CPIA_MACRO_UploadDRAM( 0, 1024, gStreamBuffer ) ) {
		tmp = printf( "UploadAndSaveDRAM: UploadDRAM failed\n" );
		goto EndUpload;
	}
	printf( "UploadAndSaveDRAM: FS uploaded\n" );
	
	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) {
		fprintf( file, " %x", (BYTE )gStreamBuffer[pos] );
		if ( !(pos%1024) ) printf( "." );
	}
	fclose( file );
	printf( "\nUploadAndSaveDRAM: FS saved\n" );


  	file = fopen( rlb_filename, "wb" );
	if ( file == NULL ) {
		printf( "UploadAndSaveDRAM: Unable to open %s file\n", rlb_filename );
		return;
	}
	printf( "UploadAndSaveDRAM: RLB save file %s\n", rlb_filename );

	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) gStreamBuffer[ pos ] = 0xFE;
	if ( !CPIA_MACRO_UploadDRAM( 1024, 2048, gStreamBuffer ) ) {
		printf( "UploadAndSaveDRAM: UploadDRAM failed\n" );
		goto EndUpload;
	}
	printf( "UploadAndSaveDRAM: RLB uploaded\n" );

	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) {
		fprintf( file, " %x", (BYTE )gStreamBuffer[pos] );
		if ( !(pos%1024) ) printf( "." );
	}
	fclose( file );

	printf( "\nUploadAndSaveDRAM: RLB saved\n" );

	return;

EndUpload:
	fclose( file );
	return;
}

BOOL UploadTestData( WORD DummyBlocks ) {

	FILE *file;
	int tmp;
	char filename[] = "dummy.txt";
	DWORD pos;
	BYTE val;
	int count;
	BOOL upload_ok = TRUE;

	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) gStreamBuffer[ pos ] = 0xFE;

	if ( !CPIA_MACRO_TestStream( 0, DummyBlocks, gStreamBuffer ) ) {
		tmp = printf( "UploadTestData: CPIA_TestStream() failed\n" );
	}
/*
	file = fopen( filename, "wb" );
	if ( file == NULL ) {
		printf( "UploadBlock: Unable to open %s file\n", filename );
		return;
	}
	for( pos = 0; pos < (DWORD)256*DummyBlocks; pos++ ) {
		fprintf( file, " %x", (BYTE )gStreamBuffer[pos] );
		if ( !(pos%1024) ) printf( "." );
	}
	fclose( file );
*/
	val = 0; count = 0;
	for( pos = 0; pos < ((DWORD)DummyBlocks * 256); pos++ ) {
		if ( gStreamBuffer[ pos ] != val ) {
			printf( "UploadTestData: gStreamBuffer at %lu   %d != %d\n", pos, gStreamBuffer[ pos ], val );
			count++;
			if ( count > 4 ) break;
			upload_ok = FALSE;
		}
		val++;
	}

	return( upload_ok );
}


void DoUploadImage( void ) {

	FILE *file;
	int tmp;
	char filename[] = "image.txt";
	DWORD pos;
	DWORD actual_bytes;

	for( pos = 0; pos < STREAM_BUFFER_LEN; pos++ ) gStreamBuffer[ pos ] = 0xFE;

	actual_bytes = 0;
	if ( !CPIA_UploadStreamData( STREAM_BUFFER_LEN, gStreamBuffer, &actual_bytes, 2048 ) ) {
		tmp = printf( "UploadImage: UploadImageStream failed\n" );
	}
	else {
		printf( "UploadImage: Image uploaded %lu bytes\n", actual_bytes );
	}

	file = fopen( filename, "wb" );
	if ( file == NULL ) {
		printf( "UploadImage: Unable to open %s file\n", filename );
		return;
	}
/*
	for( pos = 0; pos < (DWORD)actual_bytes; pos++ ) {
*/
	for( pos = 0; pos < (DWORD)STREAM_BUFFER_LEN; pos++ ) {
		fprintf( file, " %x", (BYTE )gStreamBuffer[pos] );
		if ( !(pos%1024) ) printf( "." );
	}
	fclose( file );

	return;
}

#define EEPROM_SIZE		128


BOOL TestEEPROM( void ) {
	BOOL rtn, cmd;
	BYTE eeprom[ EEPROM_SIZE ];
	BYTE old_checksum;
	BYTE calc_checksum;
	int i;

	rtn = CPIA_ModifyCameraStatus( 4, 0, 0 );	/* clear the CmdError field */
	rtn &= CPIA_SerialStart();
	rtn &= CPIA_SerialWrite( 0xA0 );
	rtn &= CPIA_SerialWrite( 0x00 );
	rtn &= CPIA_SerialStop();
	rtn &= CPIA_GetCameraStatus( NULL, NULL, NULL, NULL, &cmd, NULL, NULL, NULL );
	if ( !rtn || (cmd&SYSTEM_FLAG) ) {
		return( FALSE );
	}

	rtn = CPIA_ModifyCameraStatus( 4, 0, 0 );	/* clear the CmdError field */

	for( i = 0; i < EEPROM_SIZE-1; i++ ) eeprom[ i ] = 0x80;

	rtn &= CPIA_SerialStart();
	rtn &= CPIA_SerialWrite( 0xA1 );
	for( i = 0; i < EEPROM_SIZE-1; i++ ) {
		rtn &= CPIA_SerialRead( TRUE, &eeprom[i] );
	}
	rtn &= CPIA_SerialRead( FALSE, &eeprom[i] );
	rtn &= CPIA_SerialStop();
	rtn &= CPIA_GetCameraStatus( NULL, NULL, NULL, NULL, &cmd, NULL, NULL, NULL );
	if ( !rtn || (cmd&SYSTEM_FLAG) ) return( FALSE );

	/* now check that the checksum is correct 	*/
	old_checksum = (int)eeprom[ 0x1F ];
	calc_checksum = 0;
	eeprom[ 0x1F ] = 0;
	for( i = 0; i < EEPROM_SIZE-1; i++ ) {
		calc_checksum += eeprom[i];
	}
//	printf( "\nold_checksum = %d    calc_checksum = %d\n", (int)old_checksum, (int)calc_checksum );

	if ( old_checksum != calc_checksum ) 
		return( FALSE );
	else
		return( TRUE );
}


 		
void Pause( WORD mSecs ) {
	DWORD start, stop;

  	start = GetTime();                                        
    do {
    	stop = GetTime();                                        
	} while( (WORD)(((stop - start) * 1000)/(DWORD)CLOCKS_PER_SEC) < mSecs );

	return;
}


